@@ -1,14 +1,4 @@
 //! GPIO and pin configuration
-//!
-//! ESP32 has very flexible pin assignment via the GPIO mux. It also has a separate RTC mux for
-//! low power and analog functions.
-//!
-//! To support this flexibility two sets of traits are supported:
-//! - The various embedded_hal properties
-//! - Dedicated [Pin], [InputPin], [OutputPin], [RTCInputPin] and [RTCOutputPin]
-//!
-//! The advantage of using the dedicated traits in peripherals is that the configuration of the
-//! IO can be done inside the peripheral instead of having to be done upfront.
 
 use core::marker::PhantomData;
 
@@ -22,15 +12,23 @@ use crate::adc;
 
 pub use chip::*;
 
-/// Extension trait to split a GPIO peripheral into independent pins and registers
-pub trait GpioExt {
-    /// The type to split the GPIO into
-    type Parts;
+/// A trait implemented by every pin insance
+pub trait Pin: Send {
+    type Error;
 
-    /// Splits the GPIO block into independent pins and registers
-    fn split(self) -> Self::Parts;
+    fn pin(&self) -> i32;
 }
 
+/// A marker trait designating a pin which is capable of
+/// operating as an input pin, even if its current mode
+/// might be a different one
+pub trait InputPin: Pin {}
+
+/// A marker trait designating a pin which is capable of
+/// operating as an output pin, even if its current mode
+/// might be a different one
+pub trait OutputPin: Pin {}
+
 /// Functions available on pins with pull up/down resistors
 //
 // This is split into a separate trait from OutputPin, because for pins which also connect to
@@ -51,32 +49,29 @@ pub trait Pull {
     fn set_floating(&mut self) -> Result<&mut Self, Self::Error>;
 }
 
-pub trait Pin: Send {
-    type Error;
-
-    fn pin(&self) -> i32;
-
-    fn reset(&mut self) -> Result<(), Self::Error>;
-}
-
-pub trait InputPin: Pin {}
-
-pub trait OutputPin: Pin {}
-
 pub trait RTCPin: Pin {
     fn rtc_pin(&self) -> i32;
 }
 
+/// A marker trait designating a pin which is capable of
+/// operating as an ADC pin, even if its current mode
+/// might be a different one
 pub trait ADCPin: Pin {
     fn adc_unit(&self) -> adc_unit_t;
     fn adc_channel(&self) -> adc_channel_t;
 }
 
+/// A marker trait designating a pin which is capable of
+/// operating as a DAC pin, even if its current mode
+/// might be a different one
 #[cfg(all(not(esp32c3), not(esp32s3)))]
 pub trait DACPin: Pin {
     fn dac_channel(&self) -> dac_channel_t;
 }
 
+/// A marker trait designating a pin which is capable of
+/// operating as a touch pin, even if its current mode
+/// might be a different one
 #[cfg(not(esp32c3))]
 pub trait TouchPin: Pin {
     fn touch_channel(&self) -> touch_pad_t;
@@ -92,222 +87,6 @@ pub struct Disabled;
 
 pub struct Unknown;
 
-/// Generic $GpioX pin
-pub struct GpioPin<MODE> {
-    pin: i32,
-    _mode: PhantomData<MODE>,
-}
-
-impl<MODE> GpioPin<MODE>
-where
-    MODE: Send,
-{
-    fn new(pin: i32) -> GpioPin<MODE> {
-        Self {
-            pin,
-            _mode: PhantomData,
-        }
-    }
-
-    fn get_input_level(&self) -> bool {
-        (unsafe { gpio_get_level(self.pin) } != 0)
-    }
-
-    #[cfg(not(feature = "ulp"))]
-    fn get_output_level(&self) -> bool {
-        let pin = self.pin as u32;
-
-        #[cfg(esp32c3)]
-        let is_set_high = unsafe { (*(GPIO_OUT_REG as *const u32) >> pin) & 0x01 != 0 };
-        #[cfg(not(esp32c3))]
-        let is_set_high = if pin <= 31 {
-            // GPIO0 - GPIO31
-            unsafe { (*(GPIO_OUT_REG as *const u32) >> pin) & 0x01 != 0 }
-        } else {
-            // GPIO32+
-            unsafe { (*(GPIO_OUT1_REG as *const u32) >> (pin - 32)) & 0x01 != 0 }
-        };
-
-        is_set_high
-    }
-
-    #[cfg(feature = "ulp")]
-    fn get_output_level(&self) -> bool {
-        (unsafe { gpio_get_output_level(self.pin) } != 0)
-    }
-
-    fn set_output_level(&mut self, on: bool) -> Result<(), EspError> {
-        esp_result!(unsafe { gpio_set_level(self.pin(), (on as u8).into()) }, ())
-    }
-
-    #[cfg(not(feature = "ulp"))]
-    pub fn get_drive_strength(&self) -> Result<DriveStrength, EspError> {
-        let mut cap: gpio_drive_cap_t = 0;
-
-        esp!(unsafe { gpio_get_drive_capability(self.pin(), &mut cap as *mut _) })?;
-
-        Ok(cap.into())
-    }
-
-    #[cfg(not(feature = "ulp"))]
-    pub fn set_drive_strength(&mut self, strength: DriveStrength) -> Result<(), EspError> {
-        esp!(unsafe { gpio_set_drive_capability(self.pin(), strength.into()) })?;
-
-        Ok(())
-    }
-}
-
-impl<MODE> Pin for GpioPin<MODE>
-where
-    MODE: Send,
-{
-    type Error = EspError;
-
-    fn pin(&self) -> i32
-    where
-        Self: Sized,
-    {
-        self.pin
-    }
-
-    fn reset(&mut self) -> Result<(), Self::Error> {
-        #[cfg(not(feature = "ulp"))]
-        let res = esp_result!(unsafe { gpio_reset_pin(self.pin) }, ());
-        #[cfg(feature = "ulp")]
-        let res = Ok(());
-
-        res
-    }
-}
-
-impl InputPin for GpioPin<Input> {}
-
-impl embedded_hal::digital::v2::InputPin for GpioPin<Input> {
-    type Error = EspError;
-
-    fn is_high(&self) -> Result<bool, Self::Error> {
-        Ok(self.get_input_level())
-    }
-
-    fn is_low(&self) -> Result<bool, Self::Error> {
-        Ok(!self.get_input_level())
-    }
-}
-
-impl OutputPin for GpioPin<Output> {}
-
-impl embedded_hal::digital::v2::OutputPin for GpioPin<Output> {
-    type Error = EspError;
-
-    fn set_high(&mut self) -> Result<(), Self::Error> {
-        self.set_output_level(true)
-    }
-
-    fn set_low(&mut self) -> Result<(), Self::Error> {
-        self.set_output_level(false)
-    }
-}
-
-impl embedded_hal::digital::v2::StatefulOutputPin for GpioPin<Output> {
-    fn is_set_high(&self) -> Result<bool, Self::Error> {
-        Ok(self.get_output_level())
-    }
-
-    fn is_set_low(&self) -> Result<bool, Self::Error> {
-        Ok(!self.get_output_level())
-    }
-}
-
-impl embedded_hal::digital::v2::toggleable::Default for GpioPin<Output> {}
-
-impl InputPin for GpioPin<InputOutput> {}
-
-impl embedded_hal::digital::v2::InputPin for GpioPin<InputOutput> {
-    type Error = EspError;
-
-    fn is_high(&self) -> Result<bool, Self::Error> {
-        Ok(self.get_input_level())
-    }
-
-    fn is_low(&self) -> Result<bool, Self::Error> {
-        Ok(!self.get_input_level())
-    }
-}
-
-impl OutputPin for GpioPin<InputOutput> {}
-
-impl embedded_hal::digital::v2::OutputPin for GpioPin<InputOutput> {
-    type Error = EspError;
-
-    fn set_high(&mut self) -> Result<(), Self::Error> {
-        self.set_output_level(true)
-    }
-
-    fn set_low(&mut self) -> Result<(), Self::Error> {
-        self.set_output_level(false)
-    }
-}
-
-impl embedded_hal::digital::v2::StatefulOutputPin for GpioPin<InputOutput> {
-    fn is_set_high(&self) -> Result<bool, Self::Error> {
-        Ok(self.get_output_level())
-    }
-
-    fn is_set_low(&self) -> Result<bool, Self::Error> {
-        Ok(!self.get_output_level())
-    }
-}
-
-impl embedded_hal::digital::v2::toggleable::Default for GpioPin<InputOutput> {}
-
-// Not possible with embedded-hal V0.2 (possible with V1.0)
-// because in V0.2, channel() does not take &self
-// impl<AN: Analog<ADC>, ADC: Adc> embedded_hal::adc::Channel for GpioPin<AN> {
-//     type ID = u8;
-
-//     fn channel() -> Self::ID {
-//         todo!()
-//     }
-// }
-
-/// Interrupt events
-///
-/// *Note: ESP32 has a bug (3.14), which prevents correct triggering of interrupts when
-/// multiple GPIO's are configured for edge triggering in a group (GPIO0-31 is one group,
-/// GPIO32-39 is the other group). This can be worked around by using level triggering on the
-/// GPIO with edge triggering on the CPU.*
-//
-// Value must correspond to values in the register
-#[cfg(not(feature = "ulp"))]
-#[derive(Copy, Clone)]
-pub enum Event {
-    /// Trigger on the rising edge
-    RisingEdge = 1,
-    /// Trigger on the falling edge
-    FallingEdge = 2,
-    /// Trigger on any edge
-    AnyEdge = 3,
-    /// Trigger while low level
-    LowLevel = 4,
-    /// Trigger while high level
-    HighLevel = 5,
-}
-
-/// Input mode via RTC (type state)
-pub struct RTCInput<MODE> {
-    _mode: PhantomData<MODE>,
-}
-
-/// Output mode via RTC (type state)
-pub struct RTCOutput<MODE> {
-    _mode: PhantomData<MODE>,
-}
-
-/// Input-output mode via RTC (type state)
-pub struct RTCInputOutput<MODE> {
-    _mode: PhantomData<MODE>,
-}
-
 /// Drive strength (values are approximates)
 #[cfg(not(feature = "ulp"))]
 pub enum DriveStrength {
@@ -343,25 +122,53 @@ impl From<gpio_drive_cap_t> for DriveStrength {
     }
 }
 
-macro_rules! impl_hal_input_pin {
-    ($pxi:ident: $mode:ident) => {
-        impl embedded_hal::digital::v2::InputPin for $pxi<$mode> {
-            type Error = EspError;
+macro_rules! impl_base {
+    ($pxi:ident) => {
+        #[allow(dead_code)]
+        impl<MODE> $pxi<MODE>
+        where
+            MODE: Send,
+        {
+            fn reset(&mut self) -> Result<(), EspError> {
+                #[cfg(not(feature = "ulp"))]
+                let res = esp_result!(unsafe { gpio_reset_pin(self.pin()) }, ());
+                #[cfg(feature = "ulp")]
+                let res = Ok(());
 
-            fn is_high(&self) -> Result<bool, Self::Error> {
-                Ok(unsafe { gpio_get_level($pxi::<$mode>::runtime_pin()) } != 0)
+                res
             }
 
-            fn is_low(&self) -> Result<bool, Self::Error> {
-                Ok(!self.is_high()?)
+            fn get_input_level(&self) -> bool {
+                (unsafe { gpio_get_level(self.pin()) } != 0)
+            }
+
+            #[cfg(not(feature = "ulp"))]
+            fn get_output_level(&self) -> bool {
+                let pin = self.pin() as u32;
+
+                #[cfg(esp32c3)]
+                let is_set_high = unsafe { (*(GPIO_OUT_REG as *const u32) >> pin) & 0x01 != 0 };
+                #[cfg(not(esp32c3))]
+                let is_set_high = if pin <= 31 {
+                    // GPIO0 - GPIO31
+                    unsafe { (*(GPIO_OUT_REG as *const u32) >> pin) & 0x01 != 0 }
+                } else {
+                    // GPIO32+
+                    unsafe { (*(GPIO_OUT1_REG as *const u32) >> (pin - 32)) & 0x01 != 0 }
+                };
+
+                is_set_high
+            }
+
+            #[cfg(feature = "ulp")]
+            fn get_output_level(&self) -> bool {
+                (unsafe { gpio_get_output_level(self.pin()) } != 0)
+            }
+
+            fn set_output_level(&mut self, on: bool) -> Result<(), EspError> {
+                esp_result!(unsafe { gpio_set_level(self.pin(), (on as u8).into()) }, ())
             }
-        }
-    };
-}
 
-macro_rules! impl_hal_output_pin {
-    ($pxi:ident: $mode:ident) => {
-        impl $pxi<$mode> {
             #[cfg(not(feature = "ulp"))]
             pub fn get_drive_strength(&self) -> Result<DriveStrength, EspError> {
                 let mut cap: gpio_drive_cap_t = 0;
@@ -377,64 +184,50 @@ macro_rules! impl_hal_output_pin {
 
                 Ok(())
             }
-        }
 
-        impl embedded_hal::digital::v2::OutputPin for $pxi<$mode> {
-            type Error = EspError;
+            fn set_disabled(&mut self) -> Result<(), EspError> {
+                esp!(unsafe { gpio_set_direction(self.pin(), gpio_mode_t_GPIO_MODE_DISABLE,) })?;
 
-            fn set_high(&mut self) -> Result<(), Self::Error> {
-                esp_result!(
-                    unsafe { gpio_set_level($pxi::<$mode>::runtime_pin(), 1) },
-                    ()
-                )
+                Ok(())
             }
 
-            fn set_low(&mut self) -> Result<(), Self::Error> {
-                esp_result!(
-                    unsafe { gpio_set_level($pxi::<$mode>::runtime_pin(), 0) },
-                    ()
-                )
-            }
-        }
+            fn set_input(&mut self) -> Result<(), EspError> {
+                self.reset()?;
+                esp!(unsafe { gpio_set_direction(self.pin(), gpio_mode_t_GPIO_MODE_INPUT) })?;
 
-        impl embedded_hal::digital::v2::StatefulOutputPin for $pxi<$mode> {
-            #[cfg(not(feature = "ulp"))]
-            fn is_set_high(&self) -> Result<bool, Self::Error> {
-                let pin = $pxi::<$mode>::runtime_pin() as u32;
-
-                #[cfg(esp32c3)]
-                let is_set_high = unsafe { (*(GPIO_OUT_REG as *const u32) >> pin) & 0x01 != 0 };
-                #[cfg(not(esp32c3))]
-                let is_set_high = if pin <= 31 {
-                    // GPIO0 - GPIO31
-                    unsafe { (*(GPIO_OUT_REG as *const u32) >> pin) & 0x01 != 0 }
-                } else {
-                    // GPIO32+
-                    unsafe { (*(GPIO_OUT1_REG as *const u32) >> (pin - 32)) & 0x01 != 0 }
-                };
-
-                Ok(is_set_high)
+                Ok(())
             }
 
-            #[cfg(feature = "ulp")]
-            fn is_set_high(&self) -> Result<bool, Self::Error> {
-                Ok(unsafe { gpio_get_output_level($pxi::<$mode>::runtime_pin()) } != 0)
+            fn set_input_output(&mut self) -> Result<(), EspError> {
+                self.reset()?;
+                esp!(unsafe {
+                    gpio_set_direction(self.pin(), gpio_mode_t_GPIO_MODE_INPUT_OUTPUT)
+                })?;
+
+                Ok(())
             }
 
-            fn is_set_low(&self) -> Result<bool, Self::Error> {
-                Ok(!self.is_set_high()?)
+            fn set_input_output_od(&mut self) -> Result<(), EspError> {
+                self.reset()?;
+                esp!(unsafe {
+                    gpio_set_direction(self.pin(), gpio_mode_t_GPIO_MODE_INPUT_OUTPUT_OD)
+                })?;
+
+                Ok(())
             }
-        }
 
-        impl embedded_hal::digital::v2::ToggleableOutputPin for $pxi<$mode> {
-            type Error = EspError;
+            fn set_output(&mut self) -> Result<(), EspError> {
+                self.reset()?;
+                esp!(unsafe { gpio_set_direction(self.pin(), gpio_mode_t_GPIO_MODE_OUTPUT,) })?;
 
-            fn toggle(&mut self) -> Result<(), Self::Error> {
-                if self.is_set_high()? {
-                    Ok(self.set_low()?)
-                } else {
-                    Ok(self.set_high()?)
-                }
+                Ok(())
+            }
+
+            fn set_output_od(&mut self) -> Result<(), EspError> {
+                self.reset()?;
+                esp!(unsafe { gpio_set_direction(self.pin(), gpio_mode_t_GPIO_MODE_OUTPUT_OD,) })?;
+
+                Ok(())
             }
         }
     };
@@ -447,24 +240,14 @@ macro_rules! impl_pull {
 
             fn set_pull_up(&mut self) -> Result<&mut Self, Self::Error> {
                 esp_result!(
-                    unsafe {
-                        gpio_set_pull_mode(
-                            $pxi::<$mode>::runtime_pin(),
-                            gpio_pull_mode_t_GPIO_PULLUP_ONLY,
-                        )
-                    },
+                    unsafe { gpio_set_pull_mode(self.pin(), gpio_pull_mode_t_GPIO_PULLUP_ONLY,) },
                     self
                 )
             }
 
             fn set_pull_down(&mut self) -> Result<&mut Self, Self::Error> {
                 esp_result!(
-                    unsafe {
-                        gpio_set_pull_mode(
-                            $pxi::<$mode>::runtime_pin(),
-                            gpio_pull_mode_t_GPIO_PULLDOWN_ONLY,
-                        )
-                    },
+                    unsafe { gpio_set_pull_mode(self.pin(), gpio_pull_mode_t_GPIO_PULLDOWN_ONLY,) },
                     self
                 )
             }
@@ -472,10 +255,7 @@ macro_rules! impl_pull {
             fn set_pull_up_down(&mut self) -> Result<&mut Self, Self::Error> {
                 esp_result!(
                     unsafe {
-                        gpio_set_pull_mode(
-                            $pxi::<$mode>::runtime_pin(),
-                            gpio_pull_mode_t_GPIO_PULLUP_PULLDOWN,
-                        )
+                        gpio_set_pull_mode(self.pin(), gpio_pull_mode_t_GPIO_PULLUP_PULLDOWN)
                     },
                     self
                 )
@@ -483,12 +263,7 @@ macro_rules! impl_pull {
 
             fn set_floating(&mut self) -> Result<&mut Self, Self::Error> {
                 esp_result!(
-                    unsafe {
-                        gpio_set_pull_mode(
-                            $pxi::<$mode>::runtime_pin(),
-                            gpio_pull_mode_t_GPIO_FLOATING,
-                        )
-                    },
+                    unsafe { gpio_set_pull_mode(self.pin(), gpio_pull_mode_t_GPIO_FLOATING,) },
                     self
                 )
             }
@@ -502,24 +277,37 @@ macro_rules! impl_input_base {
             _mode: PhantomData<MODE>,
         }
 
-        #[cfg(not(feature = "ulp"))]
         impl<MODE> $pxi<MODE>
         where
             MODE: Send,
         {
-            #[inline(always)]
-            fn runtime_pin() -> i32 {
-                $pin
+            /// # Safety
+            ///
+            /// Care should be taken not to instantiate a pin which is already used elsewhere
+            pub unsafe fn new() -> $pxi<Unknown> {
+                $pxi { _mode: PhantomData }
+            }
+
+            pub fn into_unknown(self) -> Result<$pxi<Unknown>, EspError> {
+                Ok($pxi { _mode: PhantomData })
+            }
+
+            pub fn into_disabled(mut self) -> Result<$pxi<Disabled>, EspError> {
+                self.set_disabled()?;
+
+                Ok($pxi { _mode: PhantomData })
+            }
+
+            pub fn into_input(mut self) -> Result<$pxi<Input>, EspError> {
+                self.set_input()?;
+
+                Ok($pxi { _mode: PhantomData })
             }
 
             /// Degrades a concrete pin (e.g. [`Gpio1`]) to a generic pin
             /// struct that can also be used with periphals.
             pub fn degrade(self) -> GpioPin<MODE> {
-                GpioPin::new($pin)
-            }
-
-            pub fn into_unknown(self) -> $pxi<Unknown> {
-                $pxi { _mode: PhantomData }
+                unsafe { GpioPin::new($pin) }
             }
         }
 
@@ -533,19 +321,11 @@ macro_rules! impl_input_base {
             fn pin(&self) -> i32 {
                 $pin
             }
-
-            fn reset(&mut self) -> Result<(), Self::Error> {
-                #[cfg(not(feature = "ulp"))]
-                let res = esp_result!(unsafe { gpio_reset_pin(self.pin()) }, ());
-                #[cfg(feature = "ulp")]
-                let res = Ok(());
-
-                res
-            }
         }
 
         impl<MODE> InputPin for $pxi<MODE> where MODE: Send {}
 
+        impl_base!($pxi);
         impl_hal_input_pin!($pxi: Input);
     };
 }
@@ -554,138 +334,47 @@ macro_rules! impl_input_base {
 macro_rules! impl_input_only {
     ($pxi:ident: $pin:expr) => {
         impl_input_base!($pxi: $pin);
-
-        impl<MODE> $pxi<MODE>
-        where
-            MODE: Send,
-        {
-            /// # Safety
-            ///
-            /// Care should be taken not to instantiate a pin which is already used elsewhere
-            pub unsafe fn new() -> $pxi<Unknown> {
-                $pxi { _mode: PhantomData }
-            }
-
-            pub fn into_disabled(self) -> Result<$pxi<Disabled>, EspError> {
-                esp_result!(
-                    unsafe {
-                        gpio_set_direction(
-                            $pxi::<MODE>::runtime_pin(),
-                            gpio_mode_t_GPIO_MODE_DISABLE,
-                        )
-                    },
-                    $pxi { _mode: PhantomData }
-                )
-            }
-
-            pub fn into_input(mut self) -> Result<$pxi<Input>, EspError> {
-                self.reset()?;
-                esp_result!(
-                    unsafe {
-                        gpio_set_direction($pxi::<MODE>::runtime_pin(), gpio_mode_t_GPIO_MODE_INPUT)
-                    },
-                    $pxi { _mode: PhantomData }
-                )
-            }
-        }
     };
 }
 
 macro_rules! impl_input_output {
     ($pxi:ident: $pin:expr) => {
         impl_input_base!($pxi: $pin);
+        impl_pull!($pxi: Input);
+        impl_pull!($pxi: InputOutput);
+        impl_hal_input_pin!($pxi: InputOutput);
 
         impl<MODE> OutputPin for $pxi<MODE> where MODE: Send {}
 
-        impl_hal_input_pin!($pxi: InputOutput);
         impl_hal_output_pin!($pxi: InputOutput);
         impl_hal_output_pin!($pxi: Output);
-        impl_pull!($pxi: Input);
-        impl_pull!($pxi: InputOutput);
 
         impl<MODE> $pxi<MODE>
         where
             MODE: Send,
         {
-            /// # Safety
-            ///
-            /// Care should be taken not to instantiate a pin which is already used elsewhere
-            pub unsafe fn new() -> $pxi<Unknown> {
-                $pxi { _mode: PhantomData }
-            }
-
-            pub fn into_disabled(self) -> Result<$pxi<Disabled>, EspError> {
-                esp_result!(
-                    unsafe {
-                        gpio_set_direction(
-                            $pxi::<MODE>::runtime_pin(),
-                            gpio_mode_t_GPIO_MODE_DISABLE,
-                        )
-                    },
-                    $pxi { _mode: PhantomData }
-                )
-            }
-
-            pub fn into_input(mut self) -> Result<$pxi<Input>, EspError> {
-                self.reset()?;
-                esp_result!(
-                    unsafe {
-                        gpio_set_direction($pxi::<MODE>::runtime_pin(), gpio_mode_t_GPIO_MODE_INPUT)
-                    },
-                    $pxi { _mode: PhantomData }
-                )
-            }
-
             pub fn into_input_output(mut self) -> Result<$pxi<InputOutput>, EspError> {
-                self.reset()?;
-                esp_result!(
-                    unsafe {
-                        gpio_set_direction(
-                            $pxi::<MODE>::runtime_pin(),
-                            gpio_mode_t_GPIO_MODE_INPUT_OUTPUT,
-                        )
-                    },
-                    $pxi { _mode: PhantomData }
-                )
+                self.set_input_output()?;
+
+                Ok($pxi { _mode: PhantomData })
             }
 
             pub fn into_input_output_od(mut self) -> Result<$pxi<InputOutput>, EspError> {
-                self.reset()?;
-                esp_result!(
-                    unsafe {
-                        gpio_set_direction(
-                            $pxi::<MODE>::runtime_pin(),
-                            gpio_mode_t_GPIO_MODE_INPUT_OUTPUT_OD,
-                        )
-                    },
-                    $pxi { _mode: PhantomData }
-                )
+                self.set_input_output_od()?;
+
+                Ok($pxi { _mode: PhantomData })
             }
 
             pub fn into_output(mut self) -> Result<$pxi<Output>, EspError> {
-                self.reset()?;
-                esp_result!(
-                    unsafe {
-                        gpio_set_direction(
-                            $pxi::<MODE>::runtime_pin(),
-                            gpio_mode_t_GPIO_MODE_OUTPUT,
-                        )
-                    },
-                    $pxi { _mode: PhantomData }
-                )
+                self.set_output()?;
+
+                Ok($pxi { _mode: PhantomData })
             }
 
             pub fn into_output_od(mut self) -> Result<$pxi<Output>, EspError> {
-                self.reset()?;
-                esp_result!(
-                    unsafe {
-                        gpio_set_direction(
-                            $pxi::<MODE>::runtime_pin(),
-                            gpio_mode_t_GPIO_MODE_OUTPUT_OD,
-                        )
-                    },
-                    $pxi { _mode: PhantomData }
-                )
+                self.set_output_od()?;
+
+                Ok($pxi { _mode: PhantomData })
             }
         }
     };
@@ -693,27 +382,6 @@ macro_rules! impl_input_output {
 
 macro_rules! impl_rtc {
     ($pxi:ident: $pin:expr, RTC: $rtc:expr) => {
-        #[cfg(feature = "ulp")]
-        impl<MODE> $pxi<MODE>
-        where
-            MODE: Send,
-        {
-            #[inline(always)]
-            fn runtime_pin() -> i32 {
-                $rtc
-            }
-
-            /// Degrades a concrete pin (e.g. [`Gpio1`]) to a generic pin
-            /// struct that can also be used with periphals.
-            pub fn degrade(self) -> GpioPin<MODE> {
-                GpioPin::new($pin)
-            }
-
-            pub fn into_unknown(self) -> $pxi<Unknown> {
-                $pxi { _mode: PhantomData }
-            }
-        }
-
         impl<MODE> RTCPin for $pxi<MODE>
         where
             MODE: Send,
@@ -784,14 +452,25 @@ macro_rules! impl_adc {
             }
         }
 
-        impl<AN> embedded_hal::adc::Channel<AN> for $pxi<AN>
+        impl<AN> embedded_hal_0_2::adc::Channel<AN> for $pxi<AN>
         where
             AN: adc::Analog<adc::ADC1> + Send,
         {
             type ID = u8;
 
             fn channel() -> Self::ID {
-                $adc as u8
+                $adc
+            }
+        }
+
+        impl<AN> embedded_hal::adc::nb::Channel<AN> for $pxi<AN>
+        where
+            AN: adc::Analog<adc::ADC1> + Send,
+        {
+            type ID = u8;
+
+            fn channel(&self) -> Self::ID {
+                $adc
             }
         }
     };
@@ -852,7 +531,7 @@ macro_rules! impl_adc {
             }
         }
 
-        impl<AN> embedded_hal::adc::Channel<AN> for $pxi<AN>
+        impl<AN> embedded_hal_0_2::adc::Channel<AN> for $pxi<AN>
         where
             AN: adc::Analog<adc::ADC2> + Send,
         {
@@ -862,6 +541,17 @@ macro_rules! impl_adc {
                 $adc as u8
             }
         }
+
+        impl<AN> embedded_hal::adc::nb::Channel<AN> for $pxi<AN>
+        where
+            AN: adc::Analog<adc::ADC2> + Send,
+        {
+            type ID = u8;
+
+            fn channel(&self) -> Self::ID {
+                adc_unit_t_ADC_UNIT_2 as u8
+            }
+        }
     };
 
     ($pxi:ident: $pin:expr, NOADC: $adc:expr) => {};
@@ -899,6 +589,98 @@ macro_rules! impl_touch {
     ($pxi:ident: $pin:expr, NOTOUCH: $touch:expr) => {};
 }
 
+macro_rules! impl_hal_input_pin {
+    ($pxi:ident: $mode:ident) => {
+        impl embedded_hal_0_2::digital::v2::InputPin for $pxi<$mode> {
+            type Error = EspError;
+
+            fn is_high(&self) -> Result<bool, Self::Error> {
+                Ok(self.get_input_level())
+            }
+
+            fn is_low(&self) -> Result<bool, Self::Error> {
+                Ok(!self.get_input_level())
+            }
+        }
+
+        impl embedded_hal::digital::blocking::InputPin for $pxi<$mode> {
+            type Error = EspError;
+
+            fn is_high(&self) -> Result<bool, Self::Error> {
+                Ok(self.get_input_level())
+            }
+
+            fn is_low(&self) -> Result<bool, Self::Error> {
+                Ok(!self.get_input_level())
+            }
+        }
+    };
+}
+
+macro_rules! impl_hal_output_pin {
+    ($pxi:ident: $mode:ident) => {
+        impl embedded_hal_0_2::digital::v2::OutputPin for $pxi<$mode> {
+            type Error = EspError;
+
+            fn set_high(&mut self) -> Result<(), Self::Error> {
+                self.set_output_level(true)
+            }
+
+            fn set_low(&mut self) -> Result<(), Self::Error> {
+                self.set_output_level(false)
+            }
+        }
+
+        impl embedded_hal::digital::blocking::OutputPin for $pxi<$mode> {
+            type Error = EspError;
+
+            fn set_high(&mut self) -> Result<(), Self::Error> {
+                self.set_output_level(true)
+            }
+
+            fn set_low(&mut self) -> Result<(), Self::Error> {
+                self.set_output_level(false)
+            }
+        }
+
+        impl embedded_hal::digital::blocking::StatefulOutputPin for $pxi<$mode> {
+            fn is_set_high(&self) -> Result<bool, Self::Error> {
+                Ok(self.get_output_level())
+            }
+
+            fn is_set_low(&self) -> Result<bool, Self::Error> {
+                Ok(!self.get_output_level())
+            }
+        }
+
+        impl embedded_hal_0_2::digital::v2::StatefulOutputPin for $pxi<$mode> {
+            fn is_set_high(&self) -> Result<bool, Self::Error> {
+                Ok(self.get_output_level())
+            }
+
+            fn is_set_low(&self) -> Result<bool, Self::Error> {
+                Ok(!self.get_output_level())
+            }
+        }
+
+        impl embedded_hal_0_2::digital::v2::ToggleableOutputPin for $pxi<$mode> {
+            type Error = EspError;
+
+            fn toggle(&mut self) -> Result<(), Self::Error> {
+                self.set_output_level(!self.get_output_level())
+            }
+        }
+
+        impl embedded_hal::digital::blocking::ToggleableOutputPin for $pxi<$mode> {
+            type Error = EspError;
+
+            fn toggle(&mut self) -> Result<(), Self::Error> {
+                self.set_output_level(!self.get_output_level())
+            }
+        }
+    };
+}
+
 macro_rules! pin {
     ($pxi:ident: $pin:expr, Input, $rtc:ident: $rtcno:expr, $adc:ident: $adcno:expr, $dac:ident: $dacno:expr, $touch:ident: $touchno:expr) => {
         impl_input_only!($pxi: $pin);
@@ -917,11 +699,61 @@ macro_rules! pin {
     };
 }
 
+/// Generic $GpioX pin
+pub struct GpioPin<MODE> {
+    pin: i32,
+    _mode: PhantomData<MODE>,
+}
+
+impl<MODE> GpioPin<MODE>
+where
+    MODE: Send,
+{
+    /// # Safety
+    ///
+    /// Care should be taken not to instantiate this Pin, if it is
+    /// already instantiated and used elsewhere, or if it is not set
+    /// already in the mode of operation which is being instantiated
+    pub unsafe fn new(pin: i32) -> GpioPin<MODE> {
+        Self {
+            pin,
+            _mode: PhantomData,
+        }
+    }
+}
+
+impl<MODE> Pin for GpioPin<MODE>
+where
+    MODE: Send,
+{
+    type Error = EspError;
+
+    fn pin(&self) -> i32
+    where
+        Self: Sized,
+    {
+        self.pin
+    }
+}
+
+impl InputPin for GpioPin<Input> {}
+
+impl OutputPin for GpioPin<Output> {}
+
+impl InputPin for GpioPin<InputOutput> {}
+
+impl OutputPin for GpioPin<InputOutput> {}
+
+impl_base!(GpioPin);
+impl_hal_input_pin!(GpioPin: Input);
+impl_hal_input_pin!(GpioPin: InputOutput);
+impl_hal_output_pin!(GpioPin: InputOutput);
+impl_hal_output_pin!(GpioPin: Output);
+
 #[cfg(esp32)]
 mod chip {
     use core::marker::PhantomData;
 
-    use embedded_hal::digital::v2::{OutputPin as _, StatefulOutputPin as _};
     #[cfg(not(feature = "ulp"))]
     use esp_idf_sys::*;
 
@@ -1038,7 +870,7 @@ mod chip {
     impl Pins {
         /// # Safety
         ///
-        /// Care should be taken not to instnatiate the Pins structure, if it is
+        /// Care should be taken not to instantiate the Pins structure, if it is
         /// already instantiated and used elsewhere
         pub unsafe fn new() -> Self {
             Self {
@@ -1101,7 +933,6 @@ mod chip {
 mod chip {
     use core::marker::PhantomData;
 
-    use embedded_hal::digital::v2::{OutputPin as _, StatefulOutputPin as _};
     #[cfg(not(feature = "ulp"))]
     use esp_idf_sys::*;
 
@@ -1261,7 +1092,7 @@ mod chip {
     impl Pins {
         /// # Safety
         ///
-        /// Care should be taken not to instnatiate the Pins structure, if it is
+        /// Care should be taken not to instantiate the Pins structure, if it is
         /// already instantiated and used elsewhere
         pub unsafe fn new() -> Self {
             Self {
@@ -1343,7 +1174,6 @@ mod chip {
 mod chip {
     use core::marker::PhantomData;
 
-    use embedded_hal::digital::v2::{OutputPin as _, StatefulOutputPin as _};
     use esp_idf_sys::*;
 
     use super::*;
